import React from 'react';
import DeckGL from '@deck.gl/react/typed';
import { FlyToInterpolator, MapView, WebMercatorViewport } from '@deck.gl/core/typed';
import { createBaseMap } from './BaseMap';
import FilterMenu from './FilterMenu';
import { GraphType, Layer } from '../mapvis.types';
import { SelectionLayer } from '@nebula.gl/layers';
import SelectedMenu from './SelectedMenu';
import SecondaryMenu from './SecondaryMenu';

type Props = {
  graph: GraphType;
  layers: Layer[];
  showFilter: boolean;
  setShowFilter: React.Dispatch<React.SetStateAction<boolean>>;
};

const INITIAL_VIEW_STATE = {
  latitude: 52.1006,
  longitude: 5.6464,
  zoom: 6,
  bearing: 0,
  pitch: 0,
};

const FLY_SPEED = 1000;

export function MapPanel({ graph, layers, showFilter, setShowFilter }: Props) {
  const [mapSize, setMapSize] = React.useState<boolean>(false);
  const [viewport, setViewport] = React.useState<Record<string, any>>(INITIAL_VIEW_STATE);
  const [hoverObject, setHoverObject] = React.useState<any>(null);
  const [selected, setSelected] = React.useState<any[]>([]);
  const [isSelecting, setIsSelecting] = React.useState<boolean>(false);
  const [selectingRectangle, setSelectingRectangle] = React.useState<boolean>(false);

  const getFittedViewport = (minLat: number, maxLat: number, minLon: number, maxLon: number) => {
    const viewportWebMercator = new WebMercatorViewport(viewport).fitBounds(
      [
        [minLon, minLat],
        [maxLon, maxLat],
      ],
      { padding: 20 },
    );
    const { zoom, longitude, latitude } = viewportWebMercator;
    return { zoom, longitude, latitude };
  };

  const flyToBoundingBox = (minLat: number, maxLat: number, minLon: number, maxLon: number, options = {}) => {
    const fittedViewport = getFittedViewport(minLat, maxLat, minLon, maxLon);
    setViewport((prevViewport) => ({
      ...prevViewport,
      ...options,
      ...fittedViewport,
      transitionDuration: FLY_SPEED,
      transitionInterpolator: new FlyToInterpolator(),
    }));
  };

  const dataLayers = React.useMemo(() => {
    return layers.map((layer: Layer) => {
      return new layer.type({
        id: `${layer.id}`,
        graph: graph,
        visible: layer.visible,
        config: layer.config,
        selected: selected,
        hoverObject: hoverObject,
        isSelecting: isSelecting,
        flyToBoundingBox: flyToBoundingBox,
      });
    });
  }, [layers, graph, selected, hoverObject]);

  const handleSelect = (info: any, event: any) => {
    const shiftPressed = event.srcEvent.shiftKey;
    setIsSelecting(shiftPressed);
    setSelected((prevSelected) => {
      if (!shiftPressed) {
        return info.object !== undefined ? [info.object] : [];
      } else {
        const selectedIndex = prevSelected.findIndex((obj) => obj === info.object);
        if (selectedIndex !== -1) {
          prevSelected.splice(selectedIndex, 1);
        } else {
          prevSelected.push(info.object);
        }
        return [...prevSelected];
      }
    });
  };

  const selectionLayer =
    selectingRectangle &&
    new (SelectionLayer as any)({
      id: 'selection',
      selectionType: 'rectangle',
      onSelect: ({ pickingInfos }: any) => {
        setSelected(pickingInfos.map((item: any) => item.object));
        setSelectingRectangle(false);
      },
      layerIds: layers.map((layer) => layer.id),
      getTentativeFillColor: () => [22, 37, 67, 100],
    });

  return (
    <div className={`flex-grow relative ${mapSize ? '!fixed w-screen h-screen z-10' : ''}`}>
      <DeckGL
        layers={[createBaseMap(), ...dataLayers, selectionLayer]}
        initialViewState={viewport}
        views={new MapView({ repeat: true, controller: true })}
        onViewStateChange={({ viewState }) => setViewport(viewState)}
        onClick={handleSelect}
        onHover={({ object }) => setHoverObject(object !== undefined ? object : null)}
      />
      {/* <SelectedMenu selected={selected} />
      <SecondaryMenu
        mapSize={mapSize}
        setMapSize={setMapSize}
        setViewport={setViewport}
        flyToBoundingBox={flyToBoundingBox}
        setSelectingRectangle={setSelectingRectangle}
        selectingRectangle={selectingRectangle}
      /> */}
      {/* {showFilter && <FilterMenu graph={graph} setShowFilter={setShowFilter} />} */}
    </div>
  );
}
