import React from 'react';
import { CompositeLayer } from 'deck.gl/typed';
import { GeoJsonLayer, LineLayer, PathLayer } from '@deck.gl/layers/typed';
import EdgeOptions from './EdgeOptions';
import * as d3 from 'd3';
import { getDistance, getProperty } from '../../../utlis';
import { BrushingExtension } from '@deck.gl/extensions/typed';
import { Edge, LayerProps } from '../../../mapvis.types';

export const EdgeLayerConfig = {
  width: {
    fixed: true,
    accessor: 'getWidth',
    value: 1,
    condition: '',
  },
  color: {
    fixed: true,
    accessor: 'getColor',
    value: [255, 0, 0],
    condition: '',
    scaleRange: [
      [255, 0, 0],
      [0, 0, 255],
    ],
  },
  brushing: {
    active: false,
    radius: 300000,
  },
};

export class EdgeLayer extends CompositeLayer<LayerProps> {
  static type = 'Edge';
  static layerOptions = EdgeLayerConfig;

  static generateLayerOptions(layer: any, updatedLayer: void, graphInfo: { [key: string]: any }, deleteLayer: void) {
    return <EdgeOptions layer={layer} updatedLayer={updatedLayer} graphInfo={graphInfo} deleteLayer={deleteLayer} />;
  }

  updateState({ props, oldProps, context, changeFlags }: { props: any; oldProps: any; context: any; changeFlags: any }) {
    const { graph, config } = props;

    if (!config.color.fixed) {
      const colorAttribute = graph.getGraphInfo().edgeAttributes[config.color.condition];
      const colorScale = this.createColorScale(config.color, colorAttribute);
      this.setState({ colorScale });
    }
  }

  createColorScale(colorProp: { [key: string]: any }, colorAttribute: { [key: string]: any }) {
    if (colorAttribute.dataType == 'boolean') {
      return d3.scaleOrdinal().domain(['0', '1']).range(colorProp.scaleRange);
    } else if (colorAttribute.dataType == 'number') {
      return d3.scaleLinear().domain(colorAttribute.values).range(colorProp.scaleRange);
    } else if (colorAttribute.dataType == 'string') {
      return d3
        .scaleOrdinal()
        .domain([...colorAttribute.values])
        .range(colorProp.scaleRange);
    } else {
      return null;
    }
  }

  getColor(edge: Edge, colorProps: { [key: string]: any }) {
    if (colorProps.fixed) {
      return colorProps.value !== undefined ? colorProps.value : colorProps.defaultValue;
    } else {
      const { condition } = colorProps;
      if (!condition) {
        return colorProps.defaultValue;
      }
      const value = getProperty(edge.attributes, condition);
      return this.state.colorScale(value);
    }
  }

  createSegments(edges: Edge[]) {
    // Generates a path between source and target nodes
    return edges.map((edge: Edge, index) => {
      const length = getDistance(edge.path[0], edge.path[1]);
      const nSegments = length * 10;

      let xscale = d3
        .scaleLinear()
        .domain([0, nSegments + 1])
        .range([edge.path[0][0], edge.path[1][0]]);

      let yscale = d3
        .scaleLinear()
        .domain([0, nSegments + 1])
        .range([edge.path[0][1], edge.path[1][1]]);

      let source = edge.path[0];
      let target = null;
      let local = [source];

      for (let j = 1; j <= nSegments; j++) {
        target = [xscale(j), yscale(j)];
        local.push(target);
        source = target;
      }

      local.push(edge.path.slice(-1)[0]);
      return { ...edge, path: local };
    });
  }

  bundleEdges(edges: Edge[]) {
    const newEdges = this.createSegments(edges);

    // TODO: implement edge bundeling...
    const lineGenerator = d3
      .line()
      .x((d) => d[0])
      .y((d) => d[1])
      .curve(d3.curveBundle.beta(0.85));

    const geojsonFeatures = newEdges.map((edge) => {
      const line: string | null = lineGenerator(edge.path);
      if (line === null) return;

      const coord_array = [];
      const inputArray = line.split(/\s|,/);

      for (let i = 0; i < inputArray.length; i += 1) {
        const values = inputArray[i].split(/[A-Za-z]/).filter((value) => value !== '');
        for (const value of values) {
          coord_array.push(parseFloat(value));
        }
      }

      const coordinatePairs = [];
      for (let i = 0; i < coord_array.length; i += 2) {
        const x = coord_array[i];
        const y = coord_array[i + 1];
        coordinatePairs.push([x, y]);
      }

      return { type: 'Feature', properties: { ...edge }, geometry: { type: 'LineString', coordinates: coordinatePairs } };
    });

    return {
      type: 'FeatureCollection',
      features: geojsonFeatures,
    };
  }

  renderLayers() {
    const { graph, config, visible } = this.props;

    const edges = this.bundleEdges(
      graph.getEdges().map((edge: Edge) => {
        return {
          ...edge,
          path: [this.props.graph.getNodeLocation(edge.from), this.props.graph.getNodeLocation(edge.to)],
        };
      }),
    );

    console.log('displayed edges', edges);

    return new GeoJsonLayer({
      data: edges,
      pickable: true,
      stroked: false,
      filled: false,
      lineWidthMinPixels: 0.5,
      extruded: true,
      lineWidthScale: 20,
      getFillColor: [160, 160, 180, 200],
      getLineColor: (d: any) => [250, 0, 0],
      getPointRadius: 100,
      getLineWidth: 1,
      getElevation: 30,
      lineCapRounded: true,
    });

    // return new PathLayer({
    //   data: edges,
    //   pickable: true,
    //   widthScale: 10,
    //   widthMinPixels: 1,
    //   hidden: visible,
    //   rounded: true,
    //   getPath: (d) => d.path,
    //   getColor: (d) => [200, 0, 0],
    //   getWidth: (d) => 2,
    // });

    // return new LineLayer({
    //   hidden: visible,
    //   data: graph.getEdges(),
    //   pickable: true,
    //   getWidth: (d) => 1,
    //   extensions: [new BrushingExtension()],
    //   brushingEnabled: config.brushing.active,
    //   brushingRadius: 30000,
    //   getSourcePosition: (d) => graph.getNodeLocation(d.from),
    //   getTargetPosition: (d) => graph.getNodeLocation(d.to),
    //   getColor: (d) => this.getColor(d, config.color),
    // });
  }
}

EdgeLayer.layerName = 'Edge';
