import React, { useEffect } from 'react';
import DeckGL from '@deck.gl/react/typed';
import { FlyToInterpolator, MapView, WebMercatorViewport } from '@deck.gl/core/typed';
import { SelectionLayer } from '@nebula.gl/layers';
import { Coordinate, Layer } from './mapvis.types';
import { VISComponentType, VisualizationPropTypes } from '../../common';
import { GraphMetaData } from '@graphpolaris/shared/lib/data-access/statistics';
import { SettingsContainer } from '../../components/config';
import { layerTypes } from './components/layers';
import { createBaseMap } from './components/BaseMap';
import { Input } from '../../..';
import { getProperty } from './utlis';

export type MapProps = {
  layer: undefined | 'node' | 'nodelink' | 'choropleth' | 'heatmap';
  lat: undefined;
  lon: undefined;
};

const configuration: MapProps = {
  layer: 'node',
  lat: undefined,
  lon: undefined,
};

const INITIAL_VIEW_STATE = {
  latitude: 52.1006,
  longitude: 5.6464,
  zoom: 6,
  bearing: 0,
  pitch: 0,
};

const FLY_SPEED = 1000;

export const MapVis = ({ data, configuration }: VisualizationPropTypes) => {
  const [layer, setLayer] = React.useState<Layer | undefined>(undefined);
  const [viewport, setViewport] = React.useState<Record<string, any>>(INITIAL_VIEW_STATE);
  const [hoverObject, setHoverObject] = React.useState<any>(null);
  const [selected, setSelected] = React.useState<any[]>([]);
  const [isSelecting, setIsSelecting] = React.useState<boolean>(false);
  const [selectingRectangle, setSelectingRectangle] = React.useState<boolean>(false);

  const getFittedViewport = (minLat: number, maxLat: number, minLon: number, maxLon: number) => {
    const viewportWebMercator = new WebMercatorViewport(viewport).fitBounds(
      [
        [minLon, minLat],
        [maxLon, maxLat],
      ],
      { padding: 20 },
    );
    const { zoom, longitude, latitude } = viewportWebMercator;
    return { zoom, longitude, latitude };
  };

  const flyToBoundingBox = (minLat: number, maxLat: number, minLon: number, maxLon: number, options = {}) => {
    const fittedViewport = getFittedViewport(minLat, maxLat, minLon, maxLon);
    setViewport((prevViewport) => ({
      ...prevViewport,
      ...options,
      ...fittedViewport,
      transitionDuration: FLY_SPEED,
      transitionInterpolator: new FlyToInterpolator(),
    }));
  };

  useEffect(() => {
    if (configuration.layer) {
      const layerType = layerTypes[configuration.layer] as any;

      setLayer({
        id: Date.now(),
        name: 'New layer',
        type: layerType,
        config: {
          ...layerType.layerOptions,
        },
        visible: true,
      });
    }
  }, [configuration.layer]);

  const coordinateLookup: { [id: string]: Coordinate } = data.nodes.reduce(
    (acc, node) => {
      const latitude = getProperty(node, configuration.lat.split('.'));
      const longitude = getProperty(node, configuration.lon.split('.'));

      if (latitude !== undefined && longitude !== undefined) {
        acc[node.id] = [latitude, longitude];
      }

      return acc;
    },
    {} as { [id: string]: Coordinate },
  );

  const dataLayer =
    layer &&
    new layer.type({
      id: `${layer.id}`,
      graph: data,
      visible: layer.visible,
      config: layer.config,
      selected: selected,
      hoverObject: hoverObject,
      isSelecting: isSelecting,
      getNodeLocation: (id: string) => coordinateLookup[id],
      flyToBoundingBox: flyToBoundingBox,
    });

  const selectionLayer =
    selectingRectangle &&
    new (SelectionLayer as any)({
      id: 'selection',
      selectionType: 'rectangle',
      onSelect: ({ pickingInfos }: any) => {
        setSelected(pickingInfos.map((item: any) => item.object));
        setSelectingRectangle(false);
      },
      layerIds: [layer?.id ? layer.id : ''],
      getTentativeFillColor: () => [22, 37, 67, 100],
    });

  const handleSelect = (info: any, event: any) => {
    const shiftPressed = event.srcEvent.shiftKey;
    setIsSelecting(shiftPressed);
    setSelected((prevSelected) => {
      if (!shiftPressed) {
        return info.object !== undefined ? [info.object] : [];
      } else {
        const selectedIndex = prevSelected.findIndex((obj) => obj === info.object);
        if (selectedIndex !== -1) {
          prevSelected.splice(selectedIndex, 1);
        } else {
          prevSelected.push(info.object);
        }
        return [...prevSelected];
      }
    });
  };

  return (
    <div className="w-full h-full flex-grow relative">
      <DeckGL
        layers={[createBaseMap(), dataLayer, selectionLayer]}
        initialViewState={viewport}
        views={new MapView({ repeat: true, controller: true })}
        onViewStateChange={({ viewState }) => setViewport(viewState)}
        onClick={handleSelect}
        onHover={({ object }) => setHoverObject(object !== undefined ? object : null)}
      />
    </div>
  );
};

const MapSettings = ({
  configuration,
  graph,
  updateSettings,
}: {
  configuration: MapProps;
  graph: GraphMetaData;
  updateSettings: (val: any) => void;
}) => {
  const attributePaths = new Set(Object.values(graph.nodes.types).flatMap((type) => Object.keys(type.attributes)));

  return (
    <SettingsContainer>
      <span className="text-xs font-semibold">Data layer</span>
      <Input
        type="dropdown"
        value={configuration.layer}
        options={['node', 'nodelink', 'choropleth', 'heatmap']}
        onChange={(val) => updateSettings({ layer: val })}
      />

      <span className="text-xs font-semibold">Location accessor (lat)</span>
      <Input
        type="dropdown"
        value={configuration.lat}
        options={[...attributePaths]}
        disabled={attributePaths.size < 1}
        onChange={(val) => updateSettings({ lat: val })}
      />

      <span className="text-xs font-semibold">Location accessor (lon)</span>
      <Input
        type="dropdown"
        value={configuration.lon}
        options={[...attributePaths]}
        disabled={attributePaths.size < 1}
        onChange={(val) => updateSettings({ lon: val })}
      />
    </SettingsContainer>
  );
};

export const MapComponent: VISComponentType = {
  displayName: 'MapVis',
  component: MapVis,
  settings: MapSettings,
  configuration: configuration,
};

export default MapComponent;
